"""
Copyright 2018 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

from unittest import mock

import copy
import datetime as dt
import pandas as pd

import gs_quant.risk as risk
from gs_quant.api.gs.risk import GsRiskApi
from gs_quant.base import Priceable
from gs_quant.common import AssetClass
from gs_quant.instrument import CommodSwap, EqForward, EqOption, FXOption, IRBasisSwap, IRSwap, IRSwaption, IRCap,\
    IRFloor
from gs_quant.markets import PricingContext
from gs_quant.session import Environment, GsSession
from gs_quant.target.risk import PricingDateAndMarketDataAsOf, RiskPosition, RiskRequestParameters

priceables = (
    CommodSwap('Electricity', '1y'),
    EqForward('GS.N', '1y'),
    EqOption('GS.N', '3m', 'ATMF', 'Call', 'European'),
    FXOption('EUR', 'USD', '1y', 'Call', strike_price='ATMF'),
    IRSwap('Pay', '10y', 'USD'),
    IRBasisSwap('10y', 'USD'),
    IRSwaption('Pay', '10y', 'USD'),
    IRCap('10y', 'EUR'),
    IRFloor('10y', 'EUR')
)


def set_session():
    from gs_quant.session import OAuth2Session
    OAuth2Session.init = mock.MagicMock(return_value=None)
    GsSession.use(Environment.QA, 'client_id', 'secret')

    import gs_quant.markets.markets as markets
    markets.close_market_date = mock.MagicMock(return_value=dt.date.today())


def structured_calc(mocker, priceable: Priceable, measure: risk.RiskMeasure):
    set_session()

    values = {
        '$type': 'RiskVector',
        'asset': [0.01, 0.015],
        'points': [
            {'type': 'IR', 'asset': 'USD', 'class_': 'Swap', 'point': '1y'},
            {'type': 'IR', 'asset': 'USD', 'class_': 'Swap', 'point': '2y'}
        ]
    }
    mocker.return_value = [[[[values]]]]

    expected = risk.sort_risk(pd.DataFrame([
        {'mkt_type': 'IR', 'mkt_asset': 'USD', 'mkt_class': 'Swap', 'mkt_point': '1y', 'value': 0.01},
        {'mkt_type': 'IR', 'mkt_asset': 'USD', 'mkt_class': 'Swap', 'mkt_point': '2y', 'value': 0.015}
    ]))

    current = PricingContext.current
    result = priceable.calc(measure)
    assert result.raw_value.equals(expected)
    risk_requests = (risk.RiskRequest(
        positions=(RiskPosition(instrument=priceable, quantity=1),),
        measures=(measure,),
        pricing_and_market_data_as_of=(PricingDateAndMarketDataAsOf(pricing_date=current.pricing_date,
                                                                    market=current.market),),
        parameters=RiskRequestParameters(raw_results=True),
        wait_for_results=True),)
    mocker.assert_called_with(risk_requests)


def scalar_calc(mocker, priceable: Priceable, measure: risk.RiskMeasure):
    set_session()
    mocker.return_value = [[[[{'$type': 'Risk', 'val': 0.01}]]]]

    current = PricingContext.current
    result = priceable.calc(measure)
    assert result == 0.01
    risk_requests = (risk.RiskRequest(
        positions=(RiskPosition(instrument=priceable, quantity=1),),
        measures=(measure,),
        pricing_and_market_data_as_of=(PricingDateAndMarketDataAsOf(pricing_date=current.pricing_date,
                                                                    market=current.market),),
        parameters=RiskRequestParameters(raw_results=True),
        wait_for_results=True),)
    mocker.assert_called_with(risk_requests)


def price(mocker, priceable: Priceable):
    set_session()
    mocker.return_value = [[[[{'$type': 'Risk', 'val': 0.01}]]]]

    current = PricingContext.current
    result = priceable.dollar_price()
    assert result == 0.01
    risk_requests = (risk.RiskRequest(
        positions=(RiskPosition(instrument=priceable, quantity=1),),
        measures=(risk.DollarPrice,),
        pricing_and_market_data_as_of=(PricingDateAndMarketDataAsOf(pricing_date=current.pricing_date,
                                                                    market=current.market),),
        parameters=RiskRequestParameters(raw_results=True),
        wait_for_results=True),)
    mocker.assert_called_with(risk_requests)


@mock.patch.object(GsRiskApi, '_exec')
def test_price(mocker):
    for priceable in priceables:
        price(mocker, priceable)


@mock.patch.object(GsRiskApi, '_exec')
def test_structured_calc(mocker):
    set_session()

    for priceable in priceables:
        if priceable.assetClass == AssetClass.Rates:
            for measure in (risk.IRDelta, risk.IRVega):
                structured_calc(mocker, priceable, measure)
        elif priceable.assetClass == AssetClass.FX:
            for measure in (risk.FXDelta, risk.FXGamma, risk.FXVega):
                structured_calc(mocker, priceable, measure)

    values = {
        '$type': 'RiskVector',
        'asset': [0.01, 0.015],
        'points': [
            {'type': 'IR', 'asset': 'USD', 'class_': 'Swap', 'point': '1y'},
            {'type': 'IR', 'asset': 'USD', 'class_': 'Swap', 'point': '2y'}
        ]
    }

    mocker.return_value = [[[[values]] * len(priceables)]]

    with PricingContext():
        delta_f = [p.calc(risk.IRDelta) for p in priceables]

    delta = risk.aggregate_risk(delta_f, threshold=0)

    assert len(delta) == 2


@mock.patch.object(GsRiskApi, '_exec')
def test_scalar_calc(mocker):
    for priceable in priceables:
        if priceable.assetClass == AssetClass.Equity:
            for measure in (risk.EqDelta, risk.EqGamma, risk.EqVega, risk.Theta):
                scalar_calc(mocker, priceable, measure)
        elif priceable.assetClass == AssetClass.Commod:
            for measure in (risk.CommodDelta, risk.CommodVega, risk.CommodTheta):
                scalar_calc(mocker, priceable, measure)


@mock.patch.object(GsRiskApi, '_exec')
def test_async_calc(mocker):
    set_session()

    mocker.return_value = [[[[{'$type': 'Risk', 'val': 0.01 * idx}] for idx in range(len(priceables))]]]

    with PricingContext():
        dollar_price_f = [p.dollar_price() for p in priceables]

    prices = tuple(f.result() for f in dollar_price_f)
    assert prices == tuple(0.01 * i for i in range(len(priceables)))


@mock.patch.object(GsRiskApi, '_exec')
def test_disjoint_priceables_measures(mocker):
    set_session()

    swap = priceables[4]
    swaption = priceables[6]

    mocker.return_value = [[[[{'$type': 'Risk', 'val': 0.01}]]]] * 2

    with PricingContext():
        swap_price_f = swap.price()
        swaption_dollar_price_f = swaption.dollar_price()

    assert swap_price_f.result() == 0.01
    assert swaption_dollar_price_f.result() == 0.01


def test_resolution():
    set_session()

    swap = copy.copy(priceables[4])

    assert 'fixed_rate' not in swap.as_dict()

    with mock.patch('gs_quant.api.gs.risk.GsRiskApi._exec') as mocker:
        mocker.return_value = [[[[{'$type': 'LegDefinition', 'fixedRate': 0.01}]]]]
        assert swap.fixed_rate == 0.01

    swap.notional_currency = 'GBP'
    assert 'fixed_rate' not in swap.as_dict()

    with mock.patch('gs_quant.api.gs.risk.GsRiskApi._exec') as mocker:
        mocker.return_value = [[[[{'$type': 'LegDefinition', 'fixedRate': 0.007}]]]]
        assert swap.fixed_rate == 0.007



"""
Copyright 2018 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

import datetime as dt

import numpy as np

from gs_quant.errors import *

DAYS_IN_YEAR = 365.25
DAYS_IN_WEEK = 7
HOURS_IN_DAY = 24
MINS_IN_HOUR = 60
SECS_IN_MIN = 60

SECS_IN_HOUR = SECS_IN_MIN * MINS_IN_HOUR
MINS_IN_DAY = MINS_IN_HOUR * HOURS_IN_DAY
SECS_IN_DAY = SECS_IN_MIN * MINS_IN_DAY

HOURS_IN_WEEK = HOURS_IN_DAY * DAYS_IN_WEEK
MINS_IN_WEEK = HOURS_IN_WEEK * MINS_IN_HOUR
SECS_IN_WEEK = MINS_IN_WEEK * SECS_IN_HOUR

SECS_IN_YEAR = SECS_IN_MIN * MINS_IN_HOUR * HOURS_IN_DAY * DAYS_IN_YEAR


class Timer:

    def __init__(self, print_on_exit=True, label='Execution'):
        self.print_on_exit = print_on_exit
        self.label = label

    def __enter__(self):
        self.start = dt.datetime.now()

    def __exit__(self, *args):
        self.elapsed = dt.datetime.now() - self.start

        if self.print_on_exit:
            print("{} took {} seconds".format(self.label, self.elapsed.seconds + self.elapsed.microseconds / 1000000))


def to_zulu_string(time: dt.datetime):
    return time.isoformat()[:-3] + 'Z'


def time_difference_as_string(
        time_delta: np.timedelta64,
        resolution: str = 'Second'
) -> str:
    times = [SECS_IN_YEAR, SECS_IN_WEEK, SECS_IN_DAY, SECS_IN_HOUR, SECS_IN_MIN, 1]
    time_strings = ['Year', 'Week', 'Day', 'Hour', 'Minute', 'Second']

    if resolution not in time_strings:
        raise MqValueError('incorrect resolution passed in "s"' % resolution)

    times_mapped = zip(times, time_strings)

    diff: int = abs(time_delta / np.timedelta64(1, 's'))
    result = ''

    for time, time_string in times_mapped:
        m = diff // time
        if m > 0:
            added = time_string
            if m != 1:
                added += 's'

            result += '%d %s ' % (m, added)
            diff -= m * time

        if time_string == resolution:
            break

    return result.strip()



"""
Copyright 2020 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

from dateutil.relativedelta import relativedelta as rdelta
from functools import reduce

from gs_quant.timeseries.helper import _create_enum
from .statistics import *

RebalFreq = _create_enum('RebalFreq', ['Daily', 'Monthly'])
ReturnType = _create_enum('ReturnType', ['excess_return'])


@plot_function
def basket(
    series: list,
    weights: list,
    costs: list = None,
    rebal_freq: RebalFreq = RebalFreq.DAILY,
    return_type: ReturnType = ReturnType.EXCESS_RETURN,
) -> pd.Series:
    """
    Calculates a basket return series.

    :param series: list of time series of instruments
    :param weights: list of weights
    :param costs: list of costs; defaults to costs of 0
    :param rebal_freq: rebalancing frequency - Daily or Monthly
    :param return_type: return type of underlying instruments - excess return supported only.
    :return: time-series of the resulting basket

    **Usage**

    Calculates a basket return series.

    **Examples**

    Generate price series and combine them in a basket

    >>> prices1 = generate_series(100)
    >>> prices2 = generate_series(100)
    >>> mybasket = basket([prices1, prices2], [0.5, 0.5])


    **See also**

    :func:`prices`
    """

    num_assets = len(series)
    costs = costs or [0] * num_assets

    if not all(isinstance(x, pd.Series) for x in series):
        raise TypeError("Input series must be of Pandas Series type.")

    if len(weights) != num_assets or len(weights) != len(costs):
        raise ValueError("Series, weights and costs must have the same length.")

    # For all inputs which are Pandas series, get the intersection of their calendars
    cal = pd.DatetimeIndex(
        reduce(
            np.intersect1d,
            (
                curve.index
                for curve in series + weights + costs
                if isinstance(curve, pd.Series)
            ),
        )
    )

    # Reindex inputs and convert to pandas dataframes
    series = pd.concat([curve.reindex(cal) for curve in series], axis=1)
    weights = pd.concat([pd.Series(w, index=cal) for w in weights], axis=1)
    costs = pd.concat([pd.Series(c, index=cal) for c in costs], axis=1)

    if rebal_freq == RebalFreq.DAILY:
        rebal_dates = cal
    else:
        # Get hypothetical monthly rebalances
        num_rebals = (cal[-1].year - cal[0].year) * 12 + cal[-1].month - cal[0].month
        rebal_dates = [cal[0] + i * rdelta(months=1) for i in range(num_rebals + 1)]
        # Convert these to actual calendar days
        rebal_dates = [d for d in rebal_dates if d < max(cal)]
        rebal_dates = [min(cal[cal >= date]) for date in rebal_dates]

    # Create Units dataframe
    units = pd.DataFrame(index=cal, columns=series.columns)
    output = pd.Series(index=cal)

    # Initialize backtest
    output.values[0] = 100
    units.values[0, ] = (
        output.values[0] * weights.values[0, ] / series.values[0, ]
    )

    # Run backtest
    prev_rebal = 0
    for i, date in enumerate(cal[1:], 1):
        # Update performance
        output.values[i] = output.values[i - 1] + np.dot(
            units.values[i - 1, ], series.values[i, ] - series.values[i - 1, ]
        )

        # Rebalance on rebal_dates
        if date in rebal_dates:
            # Compute costs
            actual_weights = (
                weights.values[prev_rebal, ] *
                (series.values[i, ] / series.values[prev_rebal, ]) *
                (output.values[prev_rebal] / output.values[i])
            )
            output.values[i] -= (
                np.dot(costs.values[i, ], np.abs(weights.values[i, ] - actual_weights)) *
                output.values[i]
            )

            # Rebalance
            units.values[i, ] = (
                output.values[i] * weights.values[i, ] / series.values[i, ]
            )
            prev_rebal = i
        else:
            units.values[i, ] = units.values[
                i - 1,
            ]

    return output



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

from gs_quant.target.common import *
import datetime
from typing import Tuple, Union
from enum import Enum
from gs_quant.base import Base, EnumBase, InstrumentBase, camel_case_translate, get_enum_value


class ChartFill(EnumBase, Enum):    
    
    """Chart Fill Type"""

    _None = 'None'
    Solid = 'Solid'
    Gradient = 'Gradient'
    
    def __repr__(self):
        return self.value


class ChartLineDrawType(EnumBase, Enum):    
    
    """Line Draw Type"""

    Area = 'Area'
    Bars = 'Bars'
    Candlesticks = 'Candlesticks'
    Lines = 'Lines'
    StepAfter = 'StepAfter'
    StepBefore = 'StepBefore'
    StepLinear = 'StepLinear'
    Volumes = 'Volumes'
    
    def __repr__(self):
        return self.value


class ChartLineType(EnumBase, Enum):    
    
    """Line Type"""

    Solid = 'Solid'
    Knotted = 'Knotted'
    Dashed = 'Dashed'
    
    def __repr__(self):
        return self.value


class ChartRegressionStrokeType(EnumBase, Enum):    
    
    """Chart Regression Stroke Type"""

    Line = 'Line'
    Dash = 'Dash'
    
    def __repr__(self):
        return self.value


class ChartRegressionType(EnumBase, Enum):    
    
    """Chart Regression Type"""

    Linear = 'Linear'
    Exponential = 'Exponential'
    PlotData = 'PlotData'
    
    def __repr__(self):
        return self.value


class ChartType(EnumBase, Enum):    
    
    """Chart Type"""

    line = 'line'
    scatter = 'scatter'
    
    def __repr__(self):
        return self.value


class ChartProperties(Base):
        
    """An object which chart properties"""

    @camel_case_translate
    def __init__(
        self,
        x: str = None,
        x_label: str = None,
        y: str = None,
        y_label: str = None,
        color: str = None,
        shape: str = None,
        name: str = None
    ):        
        super().__init__()
        self.x = x
        self.x_label = x_label
        self.y = y
        self.y_label = y_label
        self.color = color
        self.shape = shape
        self.name = name

    @property
    def x(self) -> str:
        return self.__x

    @x.setter
    def x(self, value: str):
        self._property_changed('x')
        self.__x = value        

    @property
    def x_label(self) -> str:
        return self.__x_label

    @x_label.setter
    def x_label(self, value: str):
        self._property_changed('x_label')
        self.__x_label = value        

    @property
    def y(self) -> str:
        return self.__y

    @y.setter
    def y(self, value: str):
        self._property_changed('y')
        self.__y = value        

    @property
    def y_label(self) -> str:
        return self.__y_label

    @y_label.setter
    def y_label(self, value: str):
        self._property_changed('y_label')
        self.__y_label = value        

    @property
    def color(self) -> str:
        return self.__color

    @color.setter
    def color(self, value: str):
        self._property_changed('color')
        self.__color = value        

    @property
    def shape(self) -> str:
        return self.__shape

    @shape.setter
    def shape(self, value: str):
        self._property_changed('shape')
        self.__shape = value        


class ChartShare(Base):
        
    """Share With View Entitlement Object only for Chart"""

    @camel_case_translate
    def __init__(
        self,
        guids: Tuple[str, ...] = None,
        version: int = None,
        name: str = None
    ):        
        super().__init__()
        self.guids = guids
        self.version = version
        self.name = name

    @property
    def guids(self) -> Tuple[str, ...]:
        """Array of guid"""
        return self.__guids

    @guids.setter
    def guids(self, value: Tuple[str, ...]):
        self._property_changed('guids')
        self.__guids = value        

    @property
    def version(self) -> int:
        """Chart Object Version"""
        return self.__version

    @version.setter
    def version(self, value: int):
        self._property_changed('version')
        self.__version = value        


class XAxisSettings(Base):
        
    """An object which contains all the settings for the X axis"""

    @camel_case_translate
    def __init__(
        self,
        x_axis_date_format: str = None,
        name: str = None
    ):        
        super().__init__()
        self.x_axis_date_format = x_axis_date_format
        self.name = name

    @property
    def x_axis_date_format(self) -> str:
        return self.__x_axis_date_format

    @x_axis_date_format.setter
    def x_axis_date_format(self, value: str):
        self._property_changed('x_axis_date_format')
        self.__x_axis_date_format = value        


class ChartExpression(Base):
        
    """An object which represent the single chart expression row"""

    @camel_case_translate
    def __init__(
        self,
        axis: str = None,
        color: str = None,
        delete_gap: bool = None,
        delete_weekends: bool = None,
        digits: float = None,
        disable: bool = None,
        fill: Union[ChartFill, str] = None,
        has_x_grid: bool = None,
        has_y_grid: bool = None,
        show_statistics: bool = None,
        hide: bool = None,
        label: str = None,
        line_type: Union[ChartLineType, str] = None,
        line_draw_type: Union[ChartLineDrawType, str] = None,
        line_transparency: float = None,
        line_width: float = None,
        x_label: str = None,
        y_label: str = None,
        name: str = None
    ):        
        super().__init__()
        self.axis = axis
        self.color = color
        self.delete_gap = delete_gap
        self.delete_weekends = delete_weekends
        self.digits = digits
        self.disable = disable
        self.fill = fill
        self.has_x_grid = has_x_grid
        self.has_y_grid = has_y_grid
        self.show_statistics = show_statistics
        self.hide = hide
        self.label = label
        self.line_type = line_type
        self.line_draw_type = line_draw_type
        self.line_transparency = line_transparency
        self.line_width = line_width
        self.x_label = x_label
        self.y_label = y_label
        self.name = name

    @property
    def axis(self) -> str:
        return self.__axis

    @axis.setter
    def axis(self, value: str):
        self._property_changed('axis')
        self.__axis = value        

    @property
    def color(self) -> str:
        return self.__color

    @color.setter
    def color(self, value: str):
        self._property_changed('color')
        self.__color = value        

    @property
    def delete_gap(self) -> bool:
        return self.__delete_gap

    @delete_gap.setter
    def delete_gap(self, value: bool):
        self._property_changed('delete_gap')
        self.__delete_gap = value        

    @property
    def delete_weekends(self) -> bool:
        return self.__delete_weekends

    @delete_weekends.setter
    def delete_weekends(self, value: bool):
        self._property_changed('delete_weekends')
        self.__delete_weekends = value        

    @property
    def digits(self) -> float:
        """The number of digits to appear after the decimal point"""
        return self.__digits

    @digits.setter
    def digits(self, value: float):
        self._property_changed('digits')
        self.__digits = value        

    @property
    def disable(self) -> bool:
        return self.__disable

    @disable.setter
    def disable(self, value: bool):
        self._property_changed('disable')
        self.__disable = value        

    @property
    def fill(self) -> Union[ChartFill, str]:
        """Chart Fill Type"""
        return self.__fill

    @fill.setter
    def fill(self, value: Union[ChartFill, str]):
        self._property_changed('fill')
        self.__fill = get_enum_value(ChartFill, value)        

    @property
    def has_x_grid(self) -> bool:
        return self.__has_x_grid

    @has_x_grid.setter
    def has_x_grid(self, value: bool):
        self._property_changed('has_x_grid')
        self.__has_x_grid = value        

    @property
    def has_y_grid(self) -> bool:
        return self.__has_y_grid

    @has_y_grid.setter
    def has_y_grid(self, value: bool):
        self._property_changed('has_y_grid')
        self.__has_y_grid = value        

    @property
    def show_statistics(self) -> bool:
        return self.__show_statistics

    @show_statistics.setter
    def show_statistics(self, value: bool):
        self._property_changed('show_statistics')
        self.__show_statistics = value        

    @property
    def hide(self) -> bool:
        return self.__hide

    @hide.setter
    def hide(self, value: bool):
        self._property_changed('hide')
        self.__hide = value        

    @property
    def label(self) -> str:
        return self.__label

    @label.setter
    def label(self, value: str):
        self._property_changed('label')
        self.__label = value        

    @property
    def line_type(self) -> Union[ChartLineType, str]:
        """Line Type"""
        return self.__line_type

    @line_type.setter
    def line_type(self, value: Union[ChartLineType, str]):
        self._property_changed('line_type')
        self.__line_type = get_enum_value(ChartLineType, value)        

    @property
    def line_draw_type(self) -> Union[ChartLineDrawType, str]:
        """Line Draw Type"""
        return self.__line_draw_type

    @line_draw_type.setter
    def line_draw_type(self, value: Union[ChartLineDrawType, str]):
        self._property_changed('line_draw_type')
        self.__line_draw_type = get_enum_value(ChartLineDrawType, value)        

    @property
    def line_transparency(self) -> float:
        """Stroke transparency"""
        return self.__line_transparency

    @line_transparency.setter
    def line_transparency(self, value: float):
        self._property_changed('line_transparency')
        self.__line_transparency = value        

    @property
    def line_width(self) -> float:
        """Stroke width"""
        return self.__line_width

    @line_width.setter
    def line_width(self, value: float):
        self._property_changed('line_width')
        self.__line_width = value        

    @property
    def x_label(self) -> str:
        return self.__x_label

    @x_label.setter
    def x_label(self, value: str):
        self._property_changed('x_label')
        self.__x_label = value        

    @property
    def y_label(self) -> str:
        return self.__y_label

    @y_label.setter
    def y_label(self, value: str):
        self._property_changed('y_label')
        self.__y_label = value        


class ChartRegression(Base):
        
    """An object which chart regression"""

    @camel_case_translate
    def __init__(
        self,
        type_: Union[ChartRegressionType, str] = None,
        line_data_index: int = None,
        is_visible: bool = None,
        stroke_color: str = None,
        stroke_type: Union[ChartRegressionStrokeType, str] = None,
        stroke_width: float = None,
        name: str = None
    ):        
        super().__init__()
        self.__type = get_enum_value(ChartRegressionType, type_)
        self.line_data_index = line_data_index
        self.is_visible = is_visible
        self.stroke_color = stroke_color
        self.stroke_type = stroke_type
        self.stroke_width = stroke_width
        self.name = name

    @property
    def type(self) -> Union[ChartRegressionType, str]:
        """Chart Regression Type"""
        return self.__type

    @type.setter
    def type(self, value: Union[ChartRegressionType, str]):
        self._property_changed('type')
        self.__type = get_enum_value(ChartRegressionType, value)        

    @property
    def line_data_index(self) -> int:
        """If Regression type is PlotData Type, line data index is required and it is the
           index of the data that return from the plot runner"""
        return self.__line_data_index

    @line_data_index.setter
    def line_data_index(self, value: int):
        self._property_changed('line_data_index')
        self.__line_data_index = value        

    @property
    def is_visible(self) -> bool:
        """Visible Flag"""
        return self.__is_visible

    @is_visible.setter
    def is_visible(self, value: bool):
        self._property_changed('is_visible')
        self.__is_visible = value        

    @property
    def stroke_color(self) -> str:
        return self.__stroke_color

    @stroke_color.setter
    def stroke_color(self, value: str):
        self._property_changed('stroke_color')
        self.__stroke_color = value        

    @property
    def stroke_type(self) -> Union[ChartRegressionStrokeType, str]:
        """Regression Stroke Type"""
        return self.__stroke_type

    @stroke_type.setter
    def stroke_type(self, value: Union[ChartRegressionStrokeType, str]):
        self._property_changed('stroke_type')
        self.__stroke_type = get_enum_value(ChartRegressionStrokeType, value)        

    @property
    def stroke_width(self) -> float:
        """Stroke Width"""
        return self.__stroke_width

    @stroke_width.setter
    def stroke_width(self, value: float):
        self._property_changed('stroke_width')
        self.__stroke_width = value        


class Chart(Base):
        
    """Object representation of a Chart"""

    @camel_case_translate
    def __init__(
        self,
        name: str,
        id_: str = None,
        owner_id: str = None,
        created_by_id: str = None,
        created_time: datetime.datetime = None,
        last_updated_by_id: str = None,
        last_updated_time: datetime.datetime = None,
        entitlements: Entitlements = None,
        entitlement_exclusions: EntitlementExclusions = None,
        folder_name: str = None,
        description: str = None,
        expressions: Tuple[ChartExpression, ...] = None,
        chart_type: Union[ChartType, str] = None,
        chart_properties: Tuple[ChartProperties, ...] = None,
        regression_properties: Tuple[ChartRegression, ...] = None,
        relative_start_date: str = None,
        relative_end_date: str = None,
        start_date: datetime.date = None,
        end_date: datetime.date = None,
        start_time: datetime.datetime = None,
        end_time: datetime.datetime = None,
        tags: Tuple[str, ...] = None,
        show_statistics: bool = None,
        copy_from_id: str = None,
        version: int = None,
        draft_view_id: str = None,
        x_axis_settings: XAxisSettings = None
    ):        
        super().__init__()
        self.__id = id_
        self.owner_id = owner_id
        self.created_by_id = created_by_id
        self.created_time = created_time
        self.last_updated_by_id = last_updated_by_id
        self.last_updated_time = last_updated_time
        self.entitlements = entitlements
        self.entitlement_exclusions = entitlement_exclusions
        self.name = name
        self.folder_name = folder_name
        self.description = description
        self.expressions = expressions
        self.chart_type = chart_type
        self.chart_properties = chart_properties
        self.regression_properties = regression_properties
        self.relative_start_date = relative_start_date
        self.relative_end_date = relative_end_date
        self.start_date = start_date
        self.end_date = end_date
        self.start_time = start_time
        self.end_time = end_time
        self.tags = tags
        self.show_statistics = show_statistics
        self.copy_from_id = copy_from_id
        self.version = version
        self.draft_view_id = draft_view_id
        self.x_axis_settings = x_axis_settings

    @property
    def id(self) -> str:
        """Marquee unique identifier"""
        return self.__id

    @id.setter
    def id(self, value: str):
        self._property_changed('id')
        self.__id = value        

    @property
    def owner_id(self) -> str:
        """Marquee unique identifier for user who owns the object."""
        return self.__owner_id

    @owner_id.setter
    def owner_id(self, value: str):
        self._property_changed('owner_id')
        self.__owner_id = value        

    @property
    def created_by_id(self) -> str:
        """Unique identifier of user who created the object."""
        return self.__created_by_id

    @created_by_id.setter
    def created_by_id(self, value: str):
        self._property_changed('created_by_id')
        self.__created_by_id = value        

    @property
    def created_time(self) -> datetime.datetime:
        """Time created. ISO 8601 formatted string."""
        return self.__created_time

    @created_time.setter
    def created_time(self, value: datetime.datetime):
        self._property_changed('created_time')
        self.__created_time = value        

    @property
    def last_updated_by_id(self) -> str:
        """Unique identifier of user who last updated the object."""
        return self.__last_updated_by_id

    @last_updated_by_id.setter
    def last_updated_by_id(self, value: str):
        self._property_changed('last_updated_by_id')
        self.__last_updated_by_id = value        

    @property
    def last_updated_time(self) -> datetime.datetime:
        """Timestamp of when the object was last updated."""
        return self.__last_updated_time

    @last_updated_time.setter
    def last_updated_time(self, value: datetime.datetime):
        self._property_changed('last_updated_time')
        self.__last_updated_time = value        

    @property
    def entitlements(self) -> Entitlements:
        """Defines the entitlements of a given resource."""
        return self.__entitlements

    @entitlements.setter
    def entitlements(self, value: Entitlements):
        self._property_changed('entitlements')
        self.__entitlements = value        

    @property
    def entitlement_exclusions(self) -> EntitlementExclusions:
        """Defines the exclusion entitlements of a given resource."""
        return self.__entitlement_exclusions

    @entitlement_exclusions.setter
    def entitlement_exclusions(self, value: EntitlementExclusions):
        self._property_changed('entitlement_exclusions')
        self.__entitlement_exclusions = value        

    @property
    def name(self) -> str:
        """Display name of the chart"""
        return self.__name

    @name.setter
    def name(self, value: str):
        self._property_changed('name')
        self.__name = value        

    @property
    def folder_name(self) -> str:
        """Folder name of the chart"""
        return self.__folder_name

    @folder_name.setter
    def folder_name(self, value: str):
        self._property_changed('folder_name')
        self.__folder_name = value        

    @property
    def description(self) -> str:
        """Free text description of a Chart"""
        return self.__description

    @description.setter
    def description(self, value: str):
        self._property_changed('description')
        self.__description = value        

    @property
    def expressions(self) -> Tuple[ChartExpression, ...]:
        """Chart expressions"""
        return self.__expressions

    @expressions.setter
    def expressions(self, value: Tuple[ChartExpression, ...]):
        self._property_changed('expressions')
        self.__expressions = value        

    @property
    def chart_type(self) -> Union[ChartType, str]:
        return self.__chart_type

    @chart_type.setter
    def chart_type(self, value: Union[ChartType, str]):
        self._property_changed('chart_type')
        self.__chart_type = get_enum_value(ChartType, value)        

    @property
    def chart_properties(self) -> Tuple[ChartProperties, ...]:
        """Chart Properties"""
        return self.__chart_properties

    @chart_properties.setter
    def chart_properties(self, value: Tuple[ChartProperties, ...]):
        self._property_changed('chart_properties')
        self.__chart_properties = value        

    @property
    def regression_properties(self) -> Tuple[ChartRegression, ...]:
        """Chart Regression Properties"""
        return self.__regression_properties

    @regression_properties.setter
    def regression_properties(self, value: Tuple[ChartRegression, ...]):
        self._property_changed('regression_properties')
        self.__regression_properties = value        

    @property
    def relative_start_date(self) -> str:
        return self.__relative_start_date

    @relative_start_date.setter
    def relative_start_date(self, value: str):
        self._property_changed('relative_start_date')
        self.__relative_start_date = value        

    @property
    def relative_end_date(self) -> str:
        return self.__relative_end_date

    @relative_end_date.setter
    def relative_end_date(self, value: str):
        self._property_changed('relative_end_date')
        self.__relative_end_date = value        

    @property
    def start_date(self) -> datetime.date:
        return self.__start_date

    @start_date.setter
    def start_date(self, value: datetime.date):
        self._property_changed('start_date')
        self.__start_date = value        

    @property
    def end_date(self) -> datetime.date:
        return self.__end_date

    @end_date.setter
    def end_date(self, value: datetime.date):
        self._property_changed('end_date')
        self.__end_date = value        

    @property
    def start_time(self) -> datetime.datetime:
        return self.__start_time

    @start_time.setter
    def start_time(self, value: datetime.datetime):
        self._property_changed('start_time')
        self.__start_time = value        

    @property
    def end_time(self) -> datetime.datetime:
        return self.__end_time

    @end_time.setter
    def end_time(self, value: datetime.datetime):
        self._property_changed('end_time')
        self.__end_time = value        

    @property
    def tags(self) -> Tuple[str, ...]:
        """Metadata associated with the object. Provide an array of strings which will be
           indexed for search and locating related objects"""
        return self.__tags

    @tags.setter
    def tags(self, value: Tuple[str, ...]):
        self._property_changed('tags')
        self.__tags = value        

    @property
    def show_statistics(self) -> bool:
        """Flag to show the statistic in the chart area"""
        return self.__show_statistics

    @show_statistics.setter
    def show_statistics(self, value: bool):
        self._property_changed('show_statistics')
        self.__show_statistics = value        

    @property
    def copy_from_id(self) -> str:
        """Marquee unique identifier"""
        return self.__copy_from_id

    @copy_from_id.setter
    def copy_from_id(self, value: str):
        self._property_changed('copy_from_id')
        self.__copy_from_id = value        

    @property
    def version(self) -> int:
        """Chart Object Version"""
        return self.__version

    @version.setter
    def version(self, value: int):
        self._property_changed('version')
        self.__version = value        

    @property
    def draft_view_id(self) -> str:
        """Marquee unique identifier"""
        return self.__draft_view_id

    @draft_view_id.setter
    def draft_view_id(self, value: str):
        self._property_changed('draft_view_id')
        self.__draft_view_id = value        

    @property
    def x_axis_settings(self) -> XAxisSettings:
        """Settings for the X axis"""
        return self.__x_axis_settings

    @x_axis_settings.setter
    def x_axis_settings(self, value: XAxisSettings):
        self._property_changed('x_axis_settings')
        self.__x_axis_settings = value        



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
import inspect
import logging
from collections import namedtuple
from enum import Enum, IntEnum
from functools import wraps
from typing import Optional, Union, List

import pandas as pd

from gs_quant.api.gs.data import QueryType


def _create_enum(name, members):
    return Enum(name, {n.upper(): n.lower() for n in members}, module=__name__)


def _create_int_enum(name, mappings):
    return IntEnum(name, {k.upper(): v for k, v in mappings.items()})


def _to_offset(tenor: str) -> pd.DateOffset:
    import re
    matcher = re.fullmatch('(\\d+)([dwmy])', tenor)
    if not matcher:
        raise ValueError('invalid tenor ' + tenor)

    ab = matcher.group(2)
    if ab == 'd':
        name = 'days'
    elif ab == 'w':
        name = 'weeks'
    elif ab == 'm':
        name = 'months'
    else:
        assert ab == 'y'
        name = 'years'

    kwarg = {name: int(matcher.group(1))}
    return pd.DateOffset(**kwarg)


Interpolate = _create_enum('Interpolate', ['intersect', 'step', 'nan', 'zero', 'time'])
Returns = _create_enum('Returns', ['simple', 'logarithmic', 'absolute'])
SeriesType = _create_enum('SeriesType', ['prices', 'returns'])
CurveType = _create_enum('CurveType', ['prices', 'excess_returns'])

Window = namedtuple('Window', ['w', 'r'])


def _check_window(series_length: int, window: Window):
    if series_length > 0 and isinstance(window.w, int) and isinstance(window.r, int):
        if window.w <= 0:
            raise ValueError('Window value must be greater than zero.')
        if window.r > series_length or window.r < 0:
            raise ValueError('Ramp value must be less than the length of the series and greater than zero.')


def apply_ramp(x: pd.Series, window: Window) -> pd.Series:
    _check_window(len(x), window)
    if isinstance(window.w, int) and window.w > len(x):  # does not restrict window size when it is a DataOffset
        return pd.Series([])
    if isinstance(window.r, pd.DateOffset):
        return x.loc[x.index[0] + window.r:]
    else:
        return x[window.r:]


def normalize_window(x: Union[pd.Series, pd.DataFrame], window: Union[Window, int, None], default_window: int = None) \
        -> Window:
    if default_window is None:
        default_window = len(x)

    if isinstance(window, int):
        window = Window(w=window, r=window)
    else:
        if window is None:
            window = Window(w=default_window, r=0)
        else:
            if isinstance(window.w, str):
                window = Window(_to_offset(window.w), window.r)
            if isinstance(window.r, str):
                window = Window(window.w, _to_offset(window.r))
            if window.w and window.r is None:
                window_size = window.w
                window = Window(w=window_size, r=window_size)
            elif window.w is None:
                window = Window(w=default_window, r=window.r)

    _check_window(default_window, window)
    return window


def plot_function(fn):
    # Indicates that fn should be exported to plottool as a pure function.
    fn.plot_function = True
    return fn


def plot_session_function(fn):
    fn.plot_function = True
    fn.requires_session = True
    return fn


def plot_measure(asset_class: Optional[tuple] = None, asset_type: Optional[tuple] = None,
                 dependencies: Optional[List[QueryType]] = []):
    # Indicates that fn should be exported to plottool as a member function / pseudo-measure.
    # Set category to None for no restrictions, else provide a tuple of allowed values.
    def decorator(fn):
        assert asset_class is None or isinstance(asset_class, tuple)
        assert asset_type is None or isinstance(asset_type, tuple)

        fn.plot_measure = True
        fn.asset_class = asset_class
        fn.asset_type = asset_type
        fn.dependencies = dependencies

        return fn

    return decorator


def plot_method(fn):
    # Indicates that fn should be exported to plottool as a method.
    fn.plot_method = True

    # Allows fn to accept and ignore real_time argument even if it is not defined in the signature
    @wraps(fn)
    def ignore_extra_argument(*args, **kwargs):
        if 'real_time' not in inspect.signature(fn).parameters:
            kwargs.pop('real_time', None)
        return fn(*args, **kwargs)

    return ignore_extra_argument


def log_return(logger: logging.Logger, message):
    def outer(fn):
        @wraps(fn)
        def inner(*args, **kwargs):
            response = fn(*args, **kwargs)
            logger.debug('%s: %s', message, response)
            return response

        return inner

    return outer



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

import threading

from gs_quant.errors import MqUninitialisedError

thread_local = threading.local()


def do_not_serialise(func):
    func.do_not_serialise = True
    return func


class ContextMeta(type):

    @property
    @do_not_serialise
    def __path_key(cls) -> str:
        return '{}_path'.format(cls.__name__)

    @property
    @do_not_serialise
    def __default_key(cls) -> str:
        return '{}_default'.format(cls.__name__)

    @classmethod
    def default_value(mcs) -> object:
        return None

    @property
    @do_not_serialise
    def path(cls) -> tuple:
        return getattr(thread_local, cls.__path_key, ())

    @property
    @do_not_serialise
    def current(cls):
        """
        The current instance of this context
        """
        current = next(iter(cls.path), cls.default)
        if current is None:
            raise MqUninitialisedError('{} is not initialised'.format(cls.__name__))

        return current

    @current.setter
    def current(cls, current):
        setattr(thread_local, cls.__path_key, (current,))

    @property
    @do_not_serialise
    def current_is_set(cls) -> bool:
        return bool(cls.path) or cls.default is not None

    @property
    @do_not_serialise
    def default(cls):
        default = getattr(thread_local, cls.__default_key, None)
        if default is None:
            default = cls.default_value()
            if default is not None:
                setattr(thread_local, cls.__default_key, default)

        return default

    def push(cls, context):
        setattr(thread_local, cls.__path_key, (context,) + cls.path)

    def pop(cls):
        path = cls.path
        setattr(thread_local, cls.__path_key, path[1:])
        return path[0]


class ContextBase(metaclass=ContextMeta):

    def __enter__(self):
        self._cls.push(self)
        setattr(thread_local, self.__entered_key, True)
        self._on_enter()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self._on_exit(exc_type, exc_val, exc_tb)
        finally:
            self._cls.pop()
            setattr(thread_local, self.__entered_key, False)

    @property
    @do_not_serialise
    def __entered_key(self) -> str:
        return '{}_entered'.format(id(self))

    @property
    @do_not_serialise
    def _cls(self) -> ContextMeta:
        seen = set()
        stack = [self.__class__]
        cls = None

        while stack:
            base = stack.pop()
            if ContextBase in base.__bases__ or ContextBaseWithDefault in base.__bases__:
                cls = base
                break

            if base not in seen:
                seen.add(base)
                stack.extend(b for b in base.__bases__ if issubclass(b, ContextBase))

        return cls or self.__class__

    @property
    @do_not_serialise
    def is_entered(self) -> bool:
        return getattr(thread_local, self.__entered_key, False)

    def _on_enter(self):
        pass

    def _on_exit(self, exc_type, exc_val, exc_tb):
        pass


class ContextBaseWithDefault(ContextBase):

    @classmethod
    def default_value(cls) -> object:
        return cls()


try:
    from contextlib import nullcontext
except ImportError:
    from contextlib import AbstractContextManager

    class nullcontext(AbstractContextManager):

        def __init__(self, enter_result=None):
            self.enter_result = enter_result

        def __enter__(self):
            return self.enter_result

        def __exit__(self, exc_type, exc_val, exc_tb):
            pass



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

from gs_quant.target.common import *
import datetime
from typing import Tuple, Union
from enum import Enum
from gs_quant.base import Base, EnumBase, InstrumentBase, camel_case_translate, get_enum_value


class ComponentType(EnumBase, Enum):    
    
    """Enum listing supported component types"""

    article = 'article'
    assetPlot = 'assetPlot'
    chart = 'chart'
    barChart = 'barChart'
    commentary = 'commentary'
    commentaryPromo = 'commentaryPromo'
    monitor = 'monitor'
    plot = 'plot'
    promo = 'promo'
    rates = 'rates'
    research = 'research'
    separator = 'separator'
    
    def __repr__(self):
        return self.value


class WorkspaceType(EnumBase, Enum):    
    
    """Enum listing support workspace types"""

    cashboard = 'cashboard'
    multiplot = 'multiplot'
    
    def __repr__(self):
        return self.value


class ArticleComponentParameters(Base):
        
    """Parameters provided for a article component"""

    @camel_case_translate
    def __init__(
        self,
        height: float,
        tooltip: str = None,
        commentary_channels: Tuple[str, ...] = None,
        commentary_to_desktop_link: bool = None,
        name: str = None
    ):        
        super().__init__()
        self.height = height
        self.tooltip = tooltip
        self.commentary_channels = commentary_channels
        self.commentary_to_desktop_link = commentary_to_desktop_link
        self.name = name

    @property
    def height(self) -> float:
        """Used for restricting the height in pixels of the component on a workspace"""
        return self.__height

    @height.setter
    def height(self, value: float):
        self._property_changed('height')
        self.__height = value        

    @property
    def tooltip(self) -> str:
        """Tooltip that is displayed in an info icon next to the component title."""
        return self.__tooltip

    @tooltip.setter
    def tooltip(self, value: str):
        self._property_changed('tooltip')
        self.__tooltip = value        

    @property
    def commentary_channels(self) -> Tuple[str, ...]:
        """Optional channels associated with a commentary component"""
        return self.__commentary_channels

    @commentary_channels.setter
    def commentary_channels(self, value: Tuple[str, ...]):
        self._property_changed('commentary_channels')
        self.__commentary_channels = value        

    @property
    def commentary_to_desktop_link(self) -> bool:
        """Whether or not to display a link from commentary to desktop in the header"""
        return self.__commentary_to_desktop_link

    @commentary_to_desktop_link.setter
    def commentary_to_desktop_link(self, value: bool):
        self._property_changed('commentary_to_desktop_link')
        self.__commentary_to_desktop_link = value        


class AssetPlotComponentParameters(Base):
        
    """Parameters provided for a asset plot component"""

    @camel_case_translate
    def __init__(
        self,
        height: float,
        tooltip: str = None,
        name: str = None
    ):        
        super().__init__()
        self.height = height
        self.tooltip = tooltip
        self.name = name

    @property
    def height(self) -> float:
        """Used for restricting the height in pixels of the component on a workspace"""
        return self.__height

    @height.setter
    def height(self, value: float):
        self._property_changed('height')
        self.__height = value        

    @property
    def tooltip(self) -> str:
        """Tooltip that is displayed in an info icon next to the component title."""
        return self.__tooltip

    @tooltip.setter
    def tooltip(self, value: str):
        self._property_changed('tooltip')
        self.__tooltip = value        


class BarChartComponentParameters(Base):
        
    """Parameters provided for a bar chart component"""

    @camel_case_translate
    def __init__(
        self,
        height: float,
        tooltip: str = None,
        hide_legend: bool = None,
        name: str = None
    ):        
        super().__init__()
        self.height = height
        self.tooltip = tooltip
        self.hide_legend = hide_legend
        self.name = name

    @property
    def height(self) -> float:
        """Used for restricting the height in pixels of the component on a workspace"""
        return self.__height

    @height.setter
    def height(self, value: float):
        self._property_changed('height')
        self.__height = value        

    @property
    def tooltip(self) -> str:
        """Tooltip that is displayed in an info icon next to the component title."""
        return self.__tooltip

    @tooltip.setter
    def tooltip(self, value: str):
        self._property_changed('tooltip')
        self.__tooltip = value        

    @property
    def hide_legend(self) -> bool:
        """Whether to hide a chart or plot component legend"""
        return self.__hide_legend

    @hide_legend.setter
    def hide_legend(self, value: bool):
        self._property_changed('hide_legend')
        self.__hide_legend = value        


class ChartComponentParameters(Base):
        
    """Parameters provided for a chart component"""

    @camel_case_translate
    def __init__(
        self,
        height: float,
        tooltip: str = None,
        hide_legend: bool = None,
        chart_name: str = None,
        name: str = None
    ):        
        super().__init__()
        self.height = height
        self.tooltip = tooltip
        self.hide_legend = hide_legend
        self.chart_name = chart_name
        self.name = name

    @property
    def height(self) -> float:
        """Used for restricting the height in pixels of the component on a workspace"""
        return self.__height

    @height.setter
    def height(self, value: float):
        self._property_changed('height')
        self.__height = value        

    @property
    def tooltip(self) -> str:
        """Tooltip that is displayed in an info icon next to the component title."""
        return self.__tooltip

    @tooltip.setter
    def tooltip(self, value: str):
        self._property_changed('tooltip')
        self.__tooltip = value        

    @property
    def hide_legend(self) -> bool:
        """Whether to hide a chart or plot component legend"""
        return self.__hide_legend

    @hide_legend.setter
    def hide_legend(self, value: bool):
        self._property_changed('hide_legend')
        self.__hide_legend = value        

    @property
    def chart_name(self) -> str:
        """Name of the chart, only if component type is chart"""
        return self.__chart_name

    @chart_name.setter
    def chart_name(self, value: str):
        self._property_changed('chart_name')
        self.__chart_name = value        


class CommentaryComponentParameters(Base):
        
    """Parameters provided for a commentary component"""

    @camel_case_translate
    def __init__(
        self,
        height: float,
        tooltip: str = None,
        commentary_channels: Tuple[str, ...] = None,
        commentary_to_desktop_link: bool = None,
        name: str = None
    ):        
        super().__init__()
        self.height = height
        self.tooltip = tooltip
        self.commentary_channels = commentary_channels
        self.commentary_to_desktop_link = commentary_to_desktop_link
        self.name = name

    @property
    def height(self) -> float:
        """Used for restricting the height in pixels of the component on a workspace"""
        return self.__height

    @height.setter
    def height(self, value: float):
        self._property_changed('height')
        self.__height = value        

    @property
    def tooltip(self) -> str:
        """Tooltip that is displayed in an info icon next to the component title."""
        return self.__tooltip

    @tooltip.setter
    def tooltip(self, value: str):
        self._property_changed('tooltip')
        self.__tooltip = value        

    @property
    def commentary_channels(self) -> Tuple[str, ...]:
        """Optional channels associated with a commentary component"""
        return self.__commentary_channels

    @commentary_channels.setter
    def commentary_channels(self, value: Tuple[str, ...]):
        self._property_changed('commentary_channels')
        self.__commentary_channels = value        

    @property
    def commentary_to_desktop_link(self) -> bool:
        """Whether or not to display a link from commentary to desktop in the header"""
        return self.__commentary_to_desktop_link

    @commentary_to_desktop_link.setter
    def commentary_to_desktop_link(self, value: bool):
        self._property_changed('commentary_to_desktop_link')
        self.__commentary_to_desktop_link = value        


class CommentaryPromoComponentParameters(Base):
        
    """Parameters provided for a commentary promo component"""

    @camel_case_translate
    def __init__(
        self,
        height: float,
        tooltip: str = None,
        commentary_channels: Tuple[str, ...] = None,
        transparent: bool = None,
        body: str = None,
        name: str = None
    ):        
        super().__init__()
        self.height = height
        self.tooltip = tooltip
        self.commentary_channels = commentary_channels
        self.transparent = transparent
        self.body = body
        self.name = name

    @property
    def height(self) -> float:
        """Used for restricting the height in pixels of the component on a workspace"""
        return self.__height

    @height.setter
    def height(self, value: float):
        self._property_changed('height')
        self.__height = value        

    @property
    def tooltip(self) -> str:
        """Tooltip that is displayed in an info icon next to the component title."""
        return self.__tooltip

    @tooltip.setter
    def tooltip(self, value: str):
        self._property_changed('tooltip')
        self.__tooltip = value        

    @property
    def commentary_channels(self) -> Tuple[str, ...]:
        """Optional channels associated with a commentary component"""
        return self.__commentary_channels

    @commentary_channels.setter
    def commentary_channels(self, value: Tuple[str, ...]):
        self._property_changed('commentary_channels')
        self.__commentary_channels = value        

    @property
    def transparent(self) -> bool:
        """Sets the component card to have a transparent background"""
        return self.__transparent

    @transparent.setter
    def transparent(self, value: bool):
        self._property_changed('transparent')
        self.__transparent = value        

    @property
    def body(self) -> str:
        """The body of content to be injected into the promo component type."""
        return self.__body

    @body.setter
    def body(self, value: str):
        self._property_changed('body')
        self.__body = value        


class Entitlements(Base):
        
    """Defines the entitlements of a given resource"""

    @camel_case_translate
    def __init__(
        self,
        view: Tuple[str, ...],
        edit: Tuple[str, ...],
        admin: Tuple[str, ...],
        name: str = None
    ):        
        super().__init__()
        self.view = view
        self.edit = edit
        self.admin = admin
        self.name = name

    @property
    def view(self) -> Tuple[str, ...]:
        """Permission to view the resource and its contents"""
        return self.__view

    @view.setter
    def view(self, value: Tuple[str, ...]):
        self._property_changed('view')
        self.__view = value        

    @property
    def edit(self) -> Tuple[str, ...]:
        """Permission to edit details about the resource content, excluding entitlements.
           Can also delete the resource"""
        return self.__edit

    @edit.setter
    def edit(self, value: Tuple[str, ...]):
        self._property_changed('edit')
        self.__edit = value        

    @property
    def admin(self) -> Tuple[str, ...]:
        """Permission to edit all details of the resource, including entitlements. Can also
           delete the resource"""
        return self.__admin

    @admin.setter
    def admin(self, value: Tuple[str, ...]):
        self._property_changed('admin')
        self.__admin = value        


class MonitorComponentParameters(Base):
        
    """Parameters provided for a monitor component"""

    @camel_case_translate
    def __init__(
        self,
        height: float,
        tooltip: str = None,
        name: str = None
    ):        
        super().__init__()
        self.height = height
        self.tooltip = tooltip
        self.name = name

    @property
    def height(self) -> float:
        """Used for restricting the height in pixels of the component on a workspace"""
        return self.__height

    @height.setter
    def height(self, value: float):
        self._property_changed('height')
        self.__height = value        

    @property
    def tooltip(self) -> str:
        """Tooltip that is displayed in an info icon next to the component title."""
        return self.__tooltip

    @tooltip.setter
    def tooltip(self, value: str):
        self._property_changed('tooltip')
        self.__tooltip = value        


class PlotComponentParameters(Base):
        
    """Parameters provided for a plot component"""

    @camel_case_translate
    def __init__(
        self,
        height: float,
        tooltip: str = None,
        hide_legend: bool = None,
        plot_frequency_mode: str = None,
        name: str = None
    ):        
        super().__init__()
        self.height = height
        self.tooltip = tooltip
        self.hide_legend = hide_legend
        self.plot_frequency_mode = plot_frequency_mode
        self.name = name

    @property
    def height(self) -> float:
        """Used for restricting the height in pixels of the component on a workspace"""
        return self.__height

    @height.setter
    def height(self, value: float):
        self._property_changed('height')
        self.__height = value        

    @property
    def tooltip(self) -> str:
        """Tooltip that is displayed in an info icon next to the component title."""
        return self.__tooltip

    @tooltip.setter
    def tooltip(self, value: str):
        self._property_changed('tooltip')
        self.__tooltip = value        

    @property
    def hide_legend(self) -> bool:
        """Whether to hide a chart or plot component legend"""
        return self.__hide_legend

    @hide_legend.setter
    def hide_legend(self, value: bool):
        self._property_changed('hide_legend')
        self.__hide_legend = value        

    @property
    def plot_frequency_mode(self) -> str:
        """For plot component types, set the plot frequency mode."""
        return self.__plot_frequency_mode

    @plot_frequency_mode.setter
    def plot_frequency_mode(self, value: str):
        self._property_changed('plot_frequency_mode')
        self.__plot_frequency_mode = value        


class PromoComponentParameters(Base):
        
    """Parameters provided for a promo component"""

    @camel_case_translate
    def __init__(
        self,
        height: float,
        transparent: bool = None,
        body: str = None,
        name: str = None
    ):        
        super().__init__()
        self.height = height
        self.transparent = transparent
        self.body = body
        self.name = name

    @property
    def height(self) -> float:
        """Used for restricting the height in pixels of the component on a workspace"""
        return self.__height

    @height.setter
    def height(self, value: float):
        self._property_changed('height')
        self.__height = value        

    @property
    def transparent(self) -> bool:
        """Sets the component card to have a transparent background"""
        return self.__transparent

    @transparent.setter
    def transparent(self, value: bool):
        self._property_changed('transparent')
        self.__transparent = value        

    @property
    def body(self) -> str:
        """The body of content to be injected into the promo component type."""
        return self.__body

    @body.setter
    def body(self, value: str):
        self._property_changed('body')
        self.__body = value        


class RatesComponentParameters(Base):
        
    """Parameters provided for a rates component"""

    @camel_case_translate
    def __init__(
        self,
        height: float,
        tooltip: str = None,
        name: str = None
    ):        
        super().__init__()
        self.height = height
        self.tooltip = tooltip
        self.name = name

    @property
    def height(self) -> float:
        """Used for restricting the height in pixels of the component on a workspace"""
        return self.__height

    @height.setter
    def height(self, value: float):
        self._property_changed('height')
        self.__height = value        

    @property
    def tooltip(self) -> str:
        """Tooltip that is displayed in an info icon next to the component title."""
        return self.__tooltip

    @tooltip.setter
    def tooltip(self, value: str):
        self._property_changed('tooltip')
        self.__tooltip = value        


class ResearchComponentParameters(Base):
        
    """Parameters provided for a research component"""

    @camel_case_translate
    def __init__(
        self,
        height: float,
        tooltip: str = None,
        commentary_channels: Tuple[str, ...] = None,
        commentary_to_desktop_link: bool = None,
        name: str = None
    ):        
        super().__init__()
        self.height = height
        self.tooltip = tooltip
        self.commentary_channels = commentary_channels
        self.commentary_to_desktop_link = commentary_to_desktop_link
        self.name = name

    @property
    def height(self) -> float:
        """Used for restricting the height in pixels of the component on a workspace"""
        return self.__height

    @height.setter
    def height(self, value: float):
        self._property_changed('height')
        self.__height = value        

    @property
    def tooltip(self) -> str:
        """Tooltip that is displayed in an info icon next to the component title."""
        return self.__tooltip

    @tooltip.setter
    def tooltip(self, value: str):
        self._property_changed('tooltip')
        self.__tooltip = value        

    @property
    def commentary_channels(self) -> Tuple[str, ...]:
        """Optional channels associated with a commentary component"""
        return self.__commentary_channels

    @commentary_channels.setter
    def commentary_channels(self, value: Tuple[str, ...]):
        self._property_changed('commentary_channels')
        self.__commentary_channels = value        

    @property
    def commentary_to_desktop_link(self) -> bool:
        """Whether or not to display a link from commentary to desktop in the header"""
        return self.__commentary_to_desktop_link

    @commentary_to_desktop_link.setter
    def commentary_to_desktop_link(self, value: bool):
        self._property_changed('commentary_to_desktop_link')
        self.__commentary_to_desktop_link = value        


class SeparatorComponentParameters(Base):
        
    """Parameters provided for a separator component"""

    @camel_case_translate
    def __init__(
        self,
        height: float,
        name: str = None
    ):        
        super().__init__()
        self.height = height
        self.name = name

    @property
    def height(self) -> float:
        """Used for restricting the height in pixels of the component on a workspace"""
        return self.__height

    @height.setter
    def height(self, value: float):
        self._property_changed('height')
        self.__height = value        

    @property
    def name(self) -> str:
        """Name of the component. For example a separator name"""
        return self.__name

    @name.setter
    def name(self, value: str):
        self._property_changed('name')
        self.__name = value        


class WorkspaceDate(Base):
        
    """Relative date to display at the bottom of the page."""

    @camel_case_translate
    def __init__(
        self,
        days: float,
        text: str,
        name: str = None
    ):        
        super().__init__()
        self.days = days
        self.text = text
        self.name = name

    @property
    def days(self) -> float:
        """Number of days to deduct relative to the current date. For example: 0 days
           represents today."""
        return self.__days

    @days.setter
    def days(self, value: float):
        self._property_changed('days')
        self.__days = value        

    @property
    def text(self) -> str:
        """Text to go next to the date. For example: Data as of."""
        return self.__text

    @text.setter
    def text(self, value: str):
        self._property_changed('text')
        self.__text = value        


class WorkspaceTab(Base):
        
    """Tab represents another workspace that can be linked to."""

    @camel_case_translate
    def __init__(
        self,
        id_: str,
        name: str
    ):        
        super().__init__()
        self.__id = id_
        self.name = name

    @property
    def id(self) -> str:
        """Marquee unique identifier"""
        return self.__id

    @id.setter
    def id(self, value: str):
        self._property_changed('id')
        self.__id = value        

    @property
    def name(self) -> str:
        """Name that appears on the tab."""
        return self.__name

    @name.setter
    def name(self, value: str):
        self._property_changed('name')
        self.__name = value        


class WorkspaceComponent(Base):
        
    """Parameters provided for a market workspace"""

    @camel_case_translate
    def __init__(
        self,
        id_: dict,
        type_: Union[ComponentType, str],
        parameters: dict = None,
        name: str = None
    ):        
        super().__init__()
        self.__id = id_
        self.__type = get_enum_value(ComponentType, type_)
        self.parameters = parameters
        self.name = name

    @property
    def id(self) -> dict:
        return self.__id

    @id.setter
    def id(self, value: dict):
        self._property_changed('id')
        self.__id = value        

    @property
    def type(self) -> Union[ComponentType, str]:
        """Enum listing supported component types"""
        return self.__type

    @type.setter
    def type(self, value: Union[ComponentType, str]):
        self._property_changed('type')
        self.__type = get_enum_value(ComponentType, value)        

    @property
    def parameters(self) -> dict:
        return self.__parameters

    @parameters.setter
    def parameters(self, value: dict):
        self._property_changed('parameters')
        self.__parameters = value        


class WorkspaceParameters(Base):
        
    """Parameters provided for a market workspace"""

    @camel_case_translate
    def __init__(
        self,
        layout: str,
        components: Tuple[WorkspaceComponent, ...],
        tabs: Tuple[WorkspaceTab, ...] = None,
        disclaimer: str = None,
        date: WorkspaceDate = None,
        name: str = None
    ):        
        super().__init__()
        self.layout = layout
        self.components = components
        self.tabs = tabs
        self.disclaimer = disclaimer
        self.date = date
        self.name = name

    @property
    def layout(self) -> str:
        """A workspace layout expression. For example: r(c12($0))"""
        return self.__layout

    @layout.setter
    def layout(self, value: str):
        self._property_changed('layout')
        self.__layout = value        

    @property
    def components(self) -> Tuple[WorkspaceComponent, ...]:
        """Array of workspace components"""
        return self.__components

    @components.setter
    def components(self, value: Tuple[WorkspaceComponent, ...]):
        self._property_changed('components')
        self.__components = value        

    @property
    def tabs(self) -> Tuple[WorkspaceTab, ...]:
        """Tabs of additional market workspaces that can be linked to."""
        return self.__tabs

    @tabs.setter
    def tabs(self, value: Tuple[WorkspaceTab, ...]):
        self._property_changed('tabs')
        self.__tabs = value        

    @property
    def disclaimer(self) -> str:
        """Disclaimer that applies directly to the whole workspace displayed at the bottom."""
        return self.__disclaimer

    @disclaimer.setter
    def disclaimer(self, value: str):
        self._property_changed('disclaimer')
        self.__disclaimer = value        

    @property
    def date(self) -> WorkspaceDate:
        """Relative date to display at the bottom of the page."""
        return self.__date

    @date.setter
    def date(self, value: WorkspaceDate):
        self._property_changed('date')
        self.__date = value        


class Workspace(Base):
        
    """A market workspace object"""

    @camel_case_translate
    def __init__(
        self,
        parameters: WorkspaceParameters,
        type_: Union[WorkspaceType, str],
        id_: str = None,
        alias: str = None,
        name: str = None,
        tags: Tuple[str, ...] = None,
        created_time: datetime.datetime = None,
        last_updated_time: datetime.datetime = None,
        created_by_id: str = None,
        last_updated_by_id: str = None,
        owner_id: str = None,
        entitlements: Entitlements = None,
        folder_name: str = None,
        description: str = None,
        children_ids: Tuple[str, ...] = None
    ):        
        super().__init__()
        self.__id = id_
        self.alias = alias
        self.name = name
        self.__type = get_enum_value(WorkspaceType, type_)
        self.tags = tags
        self.parameters = parameters
        self.created_time = created_time
        self.last_updated_time = last_updated_time
        self.created_by_id = created_by_id
        self.last_updated_by_id = last_updated_by_id
        self.owner_id = owner_id
        self.entitlements = entitlements
        self.folder_name = folder_name
        self.description = description
        self.children_ids = children_ids

    @property
    def id(self) -> str:
        """Marquee unique identifier"""
        return self.__id

    @id.setter
    def id(self, value: str):
        self._property_changed('id')
        self.__id = value        

    @property
    def alias(self) -> str:
        """Market workspace alias"""
        return self.__alias

    @alias.setter
    def alias(self, value: str):
        self._property_changed('alias')
        self.__alias = value        

    @property
    def name(self) -> str:
        """Market workspace name"""
        return self.__name

    @name.setter
    def name(self, value: str):
        self._property_changed('name')
        self.__name = value        

    @property
    def type(self) -> Union[WorkspaceType, str]:
        """Enum listing support workspace types"""
        return self.__type

    @type.setter
    def type(self, value: Union[WorkspaceType, str]):
        self._property_changed('type')
        self.__type = get_enum_value(WorkspaceType, value)        

    @property
    def tags(self) -> Tuple[str, ...]:
        """Array of tags"""
        return self.__tags

    @tags.setter
    def tags(self, value: Tuple[str, ...]):
        self._property_changed('tags')
        self.__tags = value        

    @property
    def parameters(self) -> WorkspaceParameters:
        """Parameters provided for a market workspace"""
        return self.__parameters

    @parameters.setter
    def parameters(self, value: WorkspaceParameters):
        self._property_changed('parameters')
        self.__parameters = value        

    @property
    def created_time(self) -> datetime.datetime:
        """Time created. ISO 8601 formatted string"""
        return self.__created_time

    @created_time.setter
    def created_time(self, value: datetime.datetime):
        self._property_changed('created_time')
        self.__created_time = value        

    @property
    def last_updated_time(self) -> datetime.datetime:
        """Timestamp of when the object was last updated"""
        return self.__last_updated_time

    @last_updated_time.setter
    def last_updated_time(self, value: datetime.datetime):
        self._property_changed('last_updated_time')
        self.__last_updated_time = value        

    @property
    def created_by_id(self) -> str:
        """Unique identifier of user who created the object."""
        return self.__created_by_id

    @created_by_id.setter
    def created_by_id(self, value: str):
        self._property_changed('created_by_id')
        self.__created_by_id = value        

    @property
    def last_updated_by_id(self) -> str:
        """Unique identifier of user who last updated the object."""
        return self.__last_updated_by_id

    @last_updated_by_id.setter
    def last_updated_by_id(self, value: str):
        self._property_changed('last_updated_by_id')
        self.__last_updated_by_id = value        

    @property
    def owner_id(self) -> str:
        """Marquee unique identifier"""
        return self.__owner_id

    @owner_id.setter
    def owner_id(self, value: str):
        self._property_changed('owner_id')
        self.__owner_id = value        

    @property
    def entitlements(self) -> Entitlements:
        """Defines the entitlements of a given resource"""
        return self.__entitlements

    @entitlements.setter
    def entitlements(self, value: Entitlements):
        self._property_changed('entitlements')
        self.__entitlements = value        

    @property
    def folder_name(self) -> str:
        """Folder name of the monitor"""
        return self.__folder_name

    @folder_name.setter
    def folder_name(self, value: str):
        self._property_changed('folder_name')
        self.__folder_name = value        

    @property
    def description(self) -> str:
        """Market workspace description"""
        return self.__description

    @description.setter
    def description(self, value: str):
        self._property_changed('description')
        self.__description = value        

    @property
    def children_ids(self) -> Tuple[str, ...]:
        """Child workspaces for navigation"""
        return self.__children_ids

    @children_ids.setter
    def children_ids(self, value: Tuple[str, ...]):
        self._property_changed('children_ids')
        self.__children_ids = value        



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

import datetime as dt
import numpy as np
import calendar as cal
from enum import Enum, IntEnum
from typing import Iterable, Optional, Tuple, Union
from gs_quant.datetime.gscalendar import GsCalendar

DateOrDates = Union[dt.date, Iterable[dt.date]]


class PaymentFrequency(IntEnum):
    """Payment frequency enumeration

    Provides an enumeration of different payment frequencies used to to discount cashflows and accrue interest

    """
    DAILY = 252
    WEEKLY = 52
    SEMI_MONTHLY = 26
    MONTHLY = 12
    SEMI_QUARTERLY = 6
    QUARTERLY = 4
    TRI_ANNUALLY = 3
    SEMI_ANNUALLY = 2
    ANNUALLY = 1


class DayCountConvention(Enum):
    """Day Count Convention enumeration

    Provides an enumeration of different day count conventions for determining how interest accrues over payment periods
    for financial securities

    """

    # Actual/360: Number of days between dates divided by 360
    ACTUAL_360 = "ACTUAL_360"

    # Actual/364: Number of days between dates divided by 364
    ACTUAL_364 = "ACTUAL_364"

    # Actual/365_25: Number of days between dates divided by 365.25
    ACTUAL_365_25 = "ACTUAL_365_25"

    # Actual/365 FIXED: Number of days between dates divided by 365
    ACTUAL_365F = "ACTUAL_365F"

    # Actual/365 LEAP: Number of days between dates divided by 365 or 366 in leap years
    ACTUAL_365L = "ACTUAL_365L"

    # ONE_ONE: Always returns a day count fraction of 1
    ONE_ONE = "ONE_ONE"


def is_business_day(dates: DateOrDates, calendars: Union[str, Tuple[str, ...]] = (), week_mask: Optional[str] = None)\
        -> Union[bool, Tuple[bool]]:
    """
    Determine whether each date in dates is a business day

    :param dates: The input date or dates
    :param calendars: Calendars to use for holidays
    :param week_mask: Which days are considered weekends (defaults to Saturday and Sunday)
    :return: True/False if dates is a single date. A tuple indicating True/False for each date if dates is an iterable

    **Examples**

    >>> import datetime as dt
    >>> is_business_day(dt.date.today())
    >>> is_business_day(dt.date(2019, 7, 4), calendars=('NYSE',))
    """
    calendar = GsCalendar.get(calendars)
    res = np.is_busday(dates, busdaycal=calendar.business_day_calendar(week_mask))
    return tuple(res) if isinstance(res, np.ndarray) else res


def business_day_offset(
        dates: DateOrDates, offsets: Union[int, Iterable[int]],
        roll: str = 'raise', calendars: Union[str, Tuple[str, ...]] = (),
        week_mask: Optional[str] = None) -> DateOrDates:
    """
    Apply offsets to the dates and move to the nearest business date

    :param dates: The input date or dates
    :param offsets: The number of days by which to adjust the dates
    :param roll: Which direction to roll, in order to get to the nearest business date
    :param calendars: Calendars to use for holidays
    :param week_mask: Which days are considered weekends (defaults to Saturday and Sunday)
    :return: A date (if dates is a single date) or tuple of dates, adjusted by the offsets

    **Examples**

    >>> import datetime as dt
    >>> prev_bus_date = business_day_offset(dt.date.today(), -1, roll='forward')
    """
    calendar = GsCalendar.get(calendars)
    res = np.busday_offset(dates, offsets, roll, busdaycal=calendar.business_day_calendar(week_mask)).astype(dt.date)
    return tuple(res) if isinstance(res, np.ndarray) else res


def prev_business_date(
        dates: DateOrDates = dt.date.today(),
        calendars: Union[str, Tuple[str, ...]] = (),
        week_mask: Optional[str] = None) -> DateOrDates:
    """
    Returns the previous business date for a given date or date series, defaulting to today.

    :param dates: The input date or dates, defaults to today
    :param calendars: Calendars to use for holidays
    :param week_mask: Which days are considered weekends (defaults to Saturday and Sunday)
    :return: A date (if dates is a single date) or tuple of dates, adjusted by the offset of one day.

    **Example**

    >>> import datetime as dt
    >>> prev_bus_date = prev_business_date()
    """
    return business_day_offset(dates, -1, roll='forward', calendars=calendars, week_mask=week_mask)


def business_day_count(begin_dates: DateOrDates, end_dates: DateOrDates, calendars: Union[str, Tuple[str, ...]] = (
), week_mask: Optional[str] = None) -> Union[int, Tuple[int]]:
    """
    Determine the number of business days between begin_dates and end_dates

    :param begin_dates: A date or collection of beginning dates
    :param end_dates: A date or collection of end dates
    :param calendars: Calendars to use for holidays
    :param week_mask: Which days are considered weekends (defaults to Saturday and Sunday)
    :return: An int or tuple of ints, representing the number of business days between begin_dates and end_dates

    **Examples**

    >>> import datetime as dt
    >>> today = dt.date.today()
    >>> bus_days = business_day_count(today, today + dt.timedelta(days=7))
    """
    calendar = GsCalendar.get(calendars)
    res = np.busday_count(begin_dates, end_dates, busdaycal=calendar.business_day_calendar(week_mask))
    return tuple(res) if isinstance(res, np.ndarray) else res


def date_range(begin: Union[int, dt.date],
               end: Union[int, dt.date],
               calendars: Union[str, Tuple[str, ...]] = (),
               week_mask: Optional[str] = None) -> Iterable[dt.date]:
    """
    Construct a range of dates

    :param begin: Beginning date or int. An int will be interpreted as the number of business days before end
        (which must be a date)
    :param end: End date or int. An int will be interpreted as the number of business days after begin
        (which must be a date)
    :param calendars: Calendars to use for holidays
    :param week_mask: Which days are considered weekends (defaults to Saturday and Sunday)
    :return: A generator of dates

    >>> import datetime as dt
    >>> today = dt.date.today()
    >>> dates = tuple(date_range(5, today))
    >>>
    >>> for date in date_range(dt.date(2019, 1, 1), dt.date(2019, 2, 1)):
    >>>     print(date)
    """
    if isinstance(begin, dt.date):
        if isinstance(end, dt.date):
            def f():
                prev = begin
                if prev > end:
                    raise ValueError('begin must be <= end')

                while prev <= end:
                    yield prev
                    prev = business_day_offset(prev, 1, calendars=calendars, week_mask=week_mask)

            return (d for d in f())
        elif isinstance(end, int):
            return (business_day_offset(begin, i, calendars=calendars, week_mask=week_mask) for i in range(end))
        else:
            raise ValueError('end must be a date or int')
    elif isinstance(begin, int):
        if isinstance(end, dt.date):
            return (business_day_offset(end, -i, roll='preceding', calendars=calendars, week_mask=week_mask)
                    for i in range(begin))
        else:
            raise ValueError('end must be a date if begin is an int')
    else:
        raise ValueError('begin must be a date or int')


def has_feb_29(start: dt.date, end: dt.date):
    """
    Determine if date range has a leap day (29Feb)

    :param start: first date
    :param end: second date

    **Usage**

    Determine if a given date range contains a leap day (Feb 29). Used for various day count convention calculations
    which alter behaviour for leap years. Start date is exclusive and end date is inclusive

    **Examples**

    Determine if a given date range contains 29Feb

    >>> start = date(2020, 1, 1)
    >>> end = date(2020, 3, 15)
    >>> has_feb_29(start, end)
    """
    feb_29 = False
    for x in range(1, (end - start).days + 1):
        date = start + dt.timedelta(days=x)
        feb_29 = feb_29 | (date.month == 2 and date.day == 29)

    return feb_29


def day_count_fraction(
        start: dt.date,         # First payment date
        end: dt.date,           # Second payment date
        convention: DayCountConvention = DayCountConvention.ACTUAL_360,
        frequency: PaymentFrequency = PaymentFrequency.MONTHLY
):
    """
    Compute day count fraction between dates

    :param start: first date
    :param end: second date
    :param convention: day count convention
    :param frequency: payment frequency of instrument
    :return: day count fraction between dates per convention

    **Usage**

    Compute day count fraction between dates, based on the value of *convention*. For more information on the available
    day count conventions, see the
    `Day Count Conventions <https://developer.gs.com/docs/gsquant/guides/Dates/1-day-count-conventions>`_ guide.

    **Examples**

    Compute day count fraction between two dates using Actual/360 convention:

    >>> start = date(2015, 11, 12)
    >>> end = date(2017, 12, 15)
    >>> day_count_fraction(start, end, DayCountConvention.ACTUAL_360)

    """

    if convention == DayCountConvention.ACTUAL_360:
        return (end - start).days / 360
    elif convention == DayCountConvention.ACTUAL_364:
        return (end - start).days / 364
    elif convention == DayCountConvention.ACTUAL_365F:
        return (end - start).days / 365
    elif convention == DayCountConvention.ACTUAL_365L:
        if frequency == PaymentFrequency.ANNUALLY:
            days_in_year = 366 if has_feb_29(start, end) else 365
        else:
            days_in_year = 366 if cal.isleap(end.year) else 365
        return (end - start).days / days_in_year
    elif convention == DayCountConvention.ACTUAL_365_25:
        return (end - start).days / 365.25
    elif convention == DayCountConvention.ONE_ONE:
        return 1
    else:
        raise ValueError('Unknown day count convention: ' + convention.value)



"""
Copyright 2018 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

from gs_quant.markets.securities import *
from gs_quant.session import *


def test_get_asset(mocker):

    marquee_id = 'MA1234567890'
    mock_response = GsAsset(AssetClass.Equity, GsAssetType.Single_Stock, 'Test Asset')

    # mock GsSession
    mocker.patch.object(
        GsSession.__class__,
        'default_value',
        return_value=GsSession.get(
            Environment.QA,
            'client_id',
            'secret'))
    mocker.patch.object(GsSession.current, '_get', return_value=mock_response)

    asset = SecurityMaster.get_asset(marquee_id, AssetIdentifier.MARQUEE_ID)

    assert asset.name == "Test Asset"
    assert asset.get_type() == AssetType.STOCK

    asset = SecurityMaster.get_asset(marquee_id, AssetIdentifier.MARQUEE_ID, as_of=dt.date.today())

    assert asset.name == "Test Asset"
    assert asset.get_type() == AssetType.STOCK

    asset = SecurityMaster.get_asset(marquee_id, AssetIdentifier.MARQUEE_ID, as_of=dt.datetime.utcnow())

    assert asset.name == "Test Asset"
    assert asset.get_type() == AssetType.STOCK

    mock_response = GsAsset(AssetClass.Equity, GsAssetType.Index, 'Test Asset')
    mocker.patch.object(GsSession.current, '_get', return_value=mock_response)

    asset = SecurityMaster.get_asset(marquee_id, AssetIdentifier.MARQUEE_ID)

    assert asset.name == "Test Asset"
    assert asset.get_type() == AssetType.INDEX

    mock_response = GsAsset(AssetClass.Equity, GsAssetType.Future, 'Test Asset')
    mocker.patch.object(GsSession.current, '_get', return_value=mock_response)

    asset = SecurityMaster.get_asset(marquee_id, AssetIdentifier.MARQUEE_ID)

    assert asset.name == "Test Asset"
    assert asset.get_type() == AssetType.FUTURE

    mock_response = GsAsset(AssetClass.Equity, GsAssetType.ETF, 'Test Asset')
    mocker.patch.object(GsSession.current, '_get', return_value=mock_response)

    asset = SecurityMaster.get_asset(marquee_id, AssetIdentifier.MARQUEE_ID)

    assert asset.name == "Test Asset"
    assert asset.get_type() == AssetType.ETF

    mock_response = GsAsset(AssetClass.Equity, GsAssetType.Custom_Basket, 'Test Asset')
    mocker.patch.object(GsSession.current, '_get', return_value=mock_response)

    asset = SecurityMaster.get_asset(marquee_id, AssetIdentifier.MARQUEE_ID)

    assert asset.name == "Test Asset"
    assert asset.get_type() == AssetType.BASKET

    mock_response = {
        'results': (GsAsset(id=marquee_id, assetClass='Equity', type='Single Stock', name='Test 1'),),
    }

    mocker.patch.object(GsSession.current, '_post', return_value=mock_response)
    asset = SecurityMaster.get_asset('GS.N', AssetIdentifier.REUTERS_ID)
    assert asset.name == "Test 1"
    assert asset.get_type() == AssetType.STOCK

    asset = SecurityMaster.get_asset('GS', AssetIdentifier.TICKER, exchange_code=ExchangeCode.NYSE)
    assert asset.name == "Test 1"
    assert asset.get_type() == AssetType.STOCK

    asset = SecurityMaster.get_asset('GS', AssetIdentifier.TICKER, asset_type=AssetType.STOCK)
    assert asset.name == "Test 1"
    assert asset.get_type() == AssetType.STOCK

    mocker.patch.object(GsSession.current, '_post', return_value={'results': ()})
    asset = SecurityMaster.get_asset(marquee_id, AssetIdentifier.REUTERS_ID)
    assert asset is None


def test_asset_identifiers(mocker):
    marquee_id = 'MA1234567890'

    mocker.patch.object(
        GsSession,
        'default_value',
        return_value=GsSession.get(
            Environment.QA,
            'client_id',
            'secret'))
    mock_response = GsAsset(AssetClass.Equity, GsAssetType.Custom_Basket, 'Test Asset')
    mocker.patch.object(GsSession.current, '_get', return_value=mock_response)

    asset = SecurityMaster.get_asset(marquee_id, AssetIdentifier.MARQUEE_ID)

    mock_response = {'xrefs': (
        {
            'startDate': '1952-01-01',
            'endDate': '2018-12-31',
            'identifiers': {
                'ric': '.GSTHHOLD',
                'bbid': 'GSTHHOLD',
                'cusip': '9EQ24FOLD',
                'ticker': 'GSTHHOLD'
            }
        },
        {
            'startDate': '2019-01-01',
            'endDate': '2952-12-31',
            'identifiers': {
                'ric': '.GSTHHVIP',
                'bbid': 'GSTHHVIP',
                'cusip': '9EQ24FPE5',
                'ticker': 'GSTHHVIP',
            }
        }
    )}

    mocker.patch.object(GsSession.current, '_get', return_value=mock_response)

    identifiers = asset.get_identifiers(dt.date.today())

    assert identifiers[AssetIdentifier.REUTERS_ID.value] == '.GSTHHVIP'
    assert identifiers[AssetIdentifier.BLOOMBERG_ID.value] == 'GSTHHVIP'
    assert identifiers[AssetIdentifier.CUSIP.value] == '9EQ24FPE5'
    assert identifiers[AssetIdentifier.TICKER.value] == 'GSTHHVIP'

    assert asset.get_identifier(AssetIdentifier.REUTERS_ID, as_of=dt.date.today()) == '.GSTHHVIP'
    assert asset.get_identifier(AssetIdentifier.BLOOMBERG_ID, as_of=dt.date.today()) == 'GSTHHVIP'
    assert asset.get_identifier(AssetIdentifier.CUSIP, as_of=dt.date.today()) == '9EQ24FPE5'
    assert asset.get_identifier(AssetIdentifier.TICKER, as_of=dt.date.today()) == 'GSTHHVIP'

    market = PricingContext(dt.date(2018, 3, 1))

    with market:
        identifiers = asset.get_identifiers()

    assert identifiers[AssetIdentifier.REUTERS_ID.value] == '.GSTHHOLD'
    assert identifiers[AssetIdentifier.BLOOMBERG_ID.value] == 'GSTHHOLD'
    assert identifiers[AssetIdentifier.CUSIP.value] == '9EQ24FOLD'
    assert identifiers[AssetIdentifier.TICKER.value] == 'GSTHHOLD'

    market = PricingContext(dt.date(2018, 3, 1))

    with market:
        identifiers = asset.get_identifiers()

    assert identifiers[AssetIdentifier.REUTERS_ID.value] == '.GSTHHOLD'
    assert identifiers[AssetIdentifier.BLOOMBERG_ID.value] == 'GSTHHOLD'
    assert identifiers[AssetIdentifier.CUSIP.value] == '9EQ24FOLD'
    assert identifiers[AssetIdentifier.TICKER.value] == 'GSTHHOLD'



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

import json
import pathlib

from gs_quant.json_encoder import JSONEncoder


def _remove_unwanted(json_text):
    json_dict = json.loads(json_text)
    if "asOfTime" in json_dict:
        del json_dict["asOfTime"]
    return json_dict


def load_json_from_resource(test_file_name, json_file_name):
    with open(pathlib.Path(__file__).parents[1] / f'resources/{test_file_name}/{json_file_name}') as json_data:
        return json.load(json_data)


def mock_request(method, path, payload, test_file_name):
    queries = {
        'assetsDataGSNWithRic':
            '{"asOfTime": "2019-05-16T21:18:18.294Z", "limit": 4, "where": {"ric": ["GS.N"]}, "fields": ["ric", "id"]}',
        'assetsDataGSNWithId':
            '{"limit": 4, "fields": ["id", "ric"], "where": {"id": ["123456MW5E27U123456"]}}',
        'assetsDataSPXWithRic':
            '{"where": {"ric": [".SPX"]}, "limit": 4, "fields": ["ric", "id"]}',
        'assetsDataSPXWithId':
            '{"limit": 4, "fields": ["id", "ric"], "where": {"id": ["456123MW5E27U123456"]}}',
        'dataQueryRic':
            '{"fields": ["adjustedTradePrice"],'
            ' "format": "MessagePack", "where": {"assetId": ["123456MW5E27U123456"]}}',
        'dataQuerySPX':
            '{"fields": ["adjustedTradePrice"], "format": "MessagePack", "where": {"assetId": ["456123MW5E27U123456"]}}'
    }
    payload = _remove_unwanted(json.dumps(payload, cls=JSONEncoder) if payload else '{}')
    if method == 'GET':
        if path == '/data/datasets/TREOD':
            return load_json_from_resource(test_file_name, 'datasets_treod_response.json')
    elif method == 'POST':
        if path == '/assets/data/query':
            if payload == _remove_unwanted(queries['assetsDataGSNWithRic']) or \
                    payload == _remove_unwanted(queries['assetsDataGSNWithId']):
                return load_json_from_resource(test_file_name, 'assets_data_query_response_gsn.json')
            elif payload == _remove_unwanted(queries['assetsDataSPXWithRic']) or \
                    payload == _remove_unwanted(queries['assetsDataSPXWithId']):
                return load_json_from_resource(test_file_name, 'assets_data_query_response_spx.json')
        elif path == '/data/TREOD/query':
            if payload == _remove_unwanted(queries['dataQueryRic']):
                return load_json_from_resource(test_file_name, 'treod_query_response_gsn.json')
            elif payload == _remove_unwanted(queries['dataQuerySPX']):
                return load_json_from_resource(test_file_name, 'treod_query_response_spx.json')
    raise Exception(f'Unhandled request. Method: {method}, Path: {path}, payload: {payload} not recognized.')



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
from abc import ABCMeta, abstractmethod
import logging
from typing import Iterable, Mapping, Optional, Tuple, Union

from gs_quant.base import RiskKey
from gs_quant.risk import ErrorValue, RiskRequest
from gs_quant.risk.result_handlers import result_handlers


_logger = logging.getLogger(__name__)


class RiskApi(metaclass=ABCMeta):

    @classmethod
    def calc_multi(cls, requests: Iterable[RiskRequest]) -> Tuple[Union[dict, str], ...]:
        return tuple(cls.calc(request) for request in requests)

    @classmethod
    @abstractmethod
    def calc(cls, request: RiskRequest) -> Union[dict, str]:
        ...

    @classmethod
    @abstractmethod
    def get_results(cls, ids_to_requests: Mapping[str, RiskRequest], timeout: Optional[int] = None)\
            -> Mapping[RiskRequest, Union[Exception, dict]]:
        ...

    @classmethod
    def _handle_results(cls, request: RiskRequest, results: Iterable) -> dict:
        formatted_results = {}

        for risk_measure, position_results in zip(request.measures, results):
            for position, date_results in zip(request.positions, position_results):
                for as_of, date_result in zip(request.pricing_and_market_data_as_of, date_results):
                    handler = result_handlers.get(date_result.get('$type'))
                    risk_key = RiskKey(
                        cls,
                        as_of.pricing_date,
                        as_of.market,
                        request.parameters,
                        request.scenario,
                        risk_measure
                    )

                    try:
                        result = handler(date_result, risk_key, position.instrument) if handler else date_result
                    except Exception as e:
                        result = ErrorValue(risk_key, str(e))
                        _logger.error(result)

                    formatted_results[(risk_key, position.instrument)] = result

        return formatted_results



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

import datetime as dt
from base64 import b64encode

from gs_quant.target.common import Entitlements
from gs_quant.target.content import GetManyContentsResponse, ContentResponse, Content


class ContentFixtures:
    _data = {
        'id': 'some-id',
        'version': 'some-version',
        'name': 'some-name',
        'entitlements': {
            'view': [
                'some-view-entitlement'
            ],
            'edit': [
                'some-edit-entitlement'
            ],
            'delete': [
                'some-delete-token'
            ],
            'admin': [
                'some-admin-token'
            ]
        },
        'createdById': 'some-created-by-id',
        'createdTime': dt.date(2019, 5, 13),
        'authors': [
            {
                'name': 'some-author-name',
                'firstName': 'some-author-first-name',
                'lastName': 'some-author-last-name',
                'id': 'some-author-id',
                'division': 'some-author-division'
            }
        ],
        'lastUpdatedTime': dt.date(2019, 5, 14),
        'content': {
            'body': b64encode(b'Hello world!'),
            'mimeType': 'text/plain',
            'encoding': 'UTF-8'
        },
        'channels': [
            'some-channel'
        ]
    }

    @classmethod
    def _get_content_response(cls):
        return ContentResponse(
            id=cls._data['id'],
            version=cls._data['version'],
            name=cls._data['name'],
            entitlements=Entitlements(
                cls._data['entitlements']['view'],
                cls._data['entitlements']['edit'],
                cls._data['entitlements']['admin'],
                cls._data['entitlements']['delete']
            ),
            created_by_id=cls._data['createdById'],
            created_time=cls._data['createdTime'],
            last_updated_time=cls._data['lastUpdatedTime'],
            channels=cls._data['channels'],
            content=Content(
                cls._data['content']['body'],
                cls._data['content']['mimeType'],
                cls._data['content']['encoding']
            )
        )

    @classmethod
    def get_many_contents_response(cls, status: int = 200, message: str = "Ok"):
        content = cls._get_content_response()
        return GetManyContentsResponse(status, message, (content,))
